#!/usr/bin/env python

import unittest
import logging
import numpy
from atooms.simulation import Simulation, Scheduler, write_thermo, write_config, target_rmsd, write, store
from atooms.system import System
from atooms.trajectory import TrajectoryXYZ
from atooms.backends.dryrun import DryRun
from atooms.core.utils import setup_logging, rmd


setup_logging(level=40)

class Test(unittest.TestCase):

    def setUp(self):
        setup_logging(level=40, update=True)

    def test_steps_float(self):
        def dummy(sim):
            pass
        s = Simulation(DryRun(), output_path=None, enable_speedometer=False)
        s.add(dummy, 20)
        s.run(steps=1e3)
        # Check that restart does not allow to change number of steps on the go
        s._restart = True
        s.run(steps=2e3)
        self.assertEqual(s.current_step, 2000)

    def test_multiple_run_calls(self):
        """
        Multiple calls to run() with varying number of steps should add up
        correctly. This was not the case in version <= 1.4.3.
        """
        # Minimal backend
        class Backend(object):

            def __init__(self):
                self.system = System()

            def run(self, steps):
                for i in range(steps):
                    pass

        backend = Backend()

        # The run_until() method should work correctly
        from atooms.simulation import Simulation
        simulation = Simulation(backend)
        simulation.run(10)
        simulation.run_until(30)
        self.assertEqual(simulation.current_step, 30)

        # The run() method called twice should also work correctly
        from atooms.simulation import Simulation
        simulation = Simulation(backend)
        simulation.run(10)
        simulation.run(20)
        self.assertEqual(simulation.current_step, 30)

    def test_target(self):
        s = Simulation(DryRun(), output_path='/tmp/test_simulation/trajectory', steps=100)
        s.run()

    def test_target_restart(self):
        f = '/tmp/test_simulation/restart/trajectory'
        s = Simulation(DryRun(), output_path=f)
        s.add(write_thermo, Scheduler(20))
        s.run(100)
        data = numpy.loadtxt(f + '.thermo', unpack=True)
        self.assertEqual(int(data[0][-1]), 100)

        s = Simulation(DryRun(), output_path=f, restart=True)
        s.add(write_thermo, Scheduler(20))
        s.run(200)
        data = numpy.loadtxt(f + '.thermo', unpack=True)
        self.assertEqual(int(data[0][-1]), 200)

    def test_config(self):
        f = '/tmp/test_simulation/config/trajectory.xyz'

        bck = DryRun(System(N=1))
        bck.timestep = 0.1

        # Accessing list elements attributes works
        s = Simulation(bck, output_path=f, enable_speedometer=False, steps=100)
        s.add(write, Scheduler(20), ['system.particle[0].position'], 'output')
        s.run()

        # Test generic writer and write_config
        setup_logging(level=40, update=True)
        s = Simulation(bck, output_path=f, enable_speedometer=True, steps=100)
        s.trajectory_class = TrajectoryXYZ
        s.add(write_config, Scheduler(20))
        s.add(write, Scheduler(20), ['current_step', 'system.cell'], 'output')
        s.run()
        import os
        self.assertTrue(os.path.exists(f))
        self.assertTrue(os.path.exists(f + '.output'))
        with TrajectoryXYZ(f) as th:
            self.assertAlmostEqual(th.timestep, bck.timestep)

    def test_target_rmsd(self):
        f = '/tmp/test_simulation/config/trajectory'
        with self.assertRaises(IndexError):
            s = Simulation(DryRun(), output_path=f, steps=100)
            s.add(target_rmsd, Scheduler(20))
            s.run()
        s = Simulation(DryRun(), output_path=f, steps=100)
        s.add(target_rmsd, Scheduler(20), 1.0)
        s.run()

    def test_target_walltime(self):
        """Check that walltime targeting works."""
        from atooms.simulation.observers import target_walltime
        f = '/tmp/test_simulation/config/trajectory'
        s = Simulation(DryRun(), output_path=f, steps=1000000000)
        s.add(target_walltime, Scheduler(20), 1.)
        s.run()
        self.assertTrue(s.wall_time() > 1.)

    def test_target_restart_fake(self):
        f = '/tmp/test_simulation/restart/trajectory'
        s = Simulation(DryRun(), output_path=f)
        # s.add(WriterThermo(), Scheduler(20))
        s.add(write_thermo, Scheduler(20))
        s.run(100)
        s.run(100)
        data = numpy.loadtxt(f + '.thermo', unpack=True)
        self.assertEqual(int(data[0][-1]), 200)

    def test_scheduler(self):
        class Simulation:
            def __init__(self):
                self.current_step = 0
        s = Scheduler(3)
        sim = Simulation()
        inext = []
        for i in range(8):
            sim.current_step = i
            inext.append(s(sim))

        self.assertEqual(inext, [3, 3, 3, 6, 6, 6, 9, 9])

    def test_block(self):
        """Test periodic block scheduling"""
        def store_list(s, db):
            db.append(s.current_step)
        db = []
        s = Simulation(DryRun(), output_path=None, steps=18)
        s.add(store_list, Scheduler(block=[1, 2, 4, 8]), db=db)
        s.run()
        self.assertEqual(db, [0, 1, 2, 4, 8, 9, 10, 12, 16, 17, 18])

    def test_steps(self):
        """Test steps scheduling"""
        def store_list(s, db):
            db.append(s.current_step)
        db = []
        s = Simulation(DryRun(), output_path=None, steps=18)
        s.add(store_list, Scheduler(steps=[1, 2, 4, 8]), db=db)
        s.run()
        self.assertEqual(db, [0, 1, 2, 4, 8])

    def test_store(self):
        """Test store"""
        db = {}
        s = Simulation(DryRun())
        s.add(store, 1, [("steps", lambda sim: sim.current_step)], db)
        s.run(5)
        self.assertEqual(db["steps"], list(range(6)))

        # Store in simulation data
        s = Simulation(DryRun())
        s.add(store, 1, [("steps", lambda sim: sim.current_step)])
        s.run(5)
        self.assertEqual(s.data["steps"], list(range(6)))

        s = Simulation(DryRun(System(N=1)))
        s.add(store, 1, ['steps', ('x_1', lambda sim: sim.system.particle[0].position[0])])
        s.run(5)
        self.assertAlmostEqual(numpy.mean(s.data["x_1"]), 0.0)

    def test_reports(self):
        """
        Test that system in Simulation tracks the one in the backend even
        when the latter is reassigned.
        """
        s = Simulation(DryRun(), output_path=None, steps=10)
        s.run()
        import atooms.simulation.reports
        atooms.simulation.reports.report_simulation(s)
        s = System(N=1)
        atooms.simulation.reports.report_system(s)
        atooms.simulation.reports.report_stats({'a': [0.0, 0.0]})

    def test_system(self):
        """
        Test that system in Simulation tracks the one in the backend even
        when the latter is reassigned.
        """
        s = Simulation(DryRun(), output_path=None, steps=10)
        s.run()
        s.backend.system = None
        self.assertTrue(s.system is s.backend.system)

    def test_composite(self):
        """
        Test that composite simulation instances (a simulation within a
        simulation object) run independent of their parent instance.

        This checks that there are no regression against the bug fixed
        in 63a7e7863.
        """
        class NewSimulation(Simulation):

            def __init__(self, sim, steps=0, output_path=None, restart=False):
                Simulation.__init__(self, DryRun(), output_path=output_path,
                                    steps=steps, restart=restart)
                self.sim = sim

            def __str__(self):
                return 'NewSimulation'

            def run_until(self, steps):
                self.sim.run()
                self.current_step = steps

        sim = Simulation(DryRun(), steps=3)
        new_sim = NewSimulation(sim, steps=1)
        new_sim.run()
        self.assertEqual(new_sim.current_step, 1)
        self.assertEqual(sim.current_step, 3)

    def test_shell_stop(self):
        from atooms.simulation import shell_stop
        f = '/tmp/test_simulation/shell/trajectory'
        s = Simulation(DryRun(), output_path=f)
        s.add(shell_stop, Scheduler(steps=[20]), 'exit 1')
        s.add(write_thermo, Scheduler(10))
        s.run(100)
        self.assertEqual(s.current_step, 20)

        s = Simulation(DryRun(), output_path=f)
        s.add(shell_stop, Scheduler(20), 'exit 0')
        s.add(write_thermo, Scheduler(10))
        s.run(100)
        self.assertEqual(s.current_step, 100)

        # Test formatted string
        s = Simulation(DryRun(), output_path=f)
        s.add(shell_stop, Scheduler(20), '[ {sim.current_step} -eq 40 ] && exit 1 || exit 0')
        s.run(100)
        self.assertEqual(s.current_step, 40)

    def test_walltime(self):
        s = Simulation(DryRun())
        s.run(100)
        wt = s.wall_time()
        self.assertEqual(s.wall_time(), wt)
        s.run(100)
        wt1 = s.wall_time()
        self.assertTrue(wt1 > wt)

    def test_python_stop(self):
        from atooms.simulation import target_python_stop
        f = '/tmp/test_simulation/python/trajectory'
        s = Simulation(DryRun(), output_path=f)
        s.add(target_python_stop, Scheduler(20), '{current_step} == 40')
        s.add(write_thermo, Scheduler(10))
        s.run(100)
        self.assertEqual(s.current_step, 40)

    def test_multiple_observers(self):
        """We can add multiple observers of the same kind"""
        from atooms.core.utils import mkdir
        from atooms.trajectory import TrajectoryXYZ, TrajectoryRam
        from atooms.simulation import write_trajectory
        mkdir('/tmp/test_simulation')
        th0 = TrajectoryRam()
        th1 = TrajectoryXYZ('/tmp/test_simulation/trajectory.xyz', 'w')
        s = Simulation(DryRun())
        s.add(write_trajectory, Scheduler(1), trajectory=th0)
        s.add(write_trajectory, Scheduler(1), trajectory=th1)
        s.run(100)
        th0.close()
        th1.close()
        self.assertEqual(s.current_step, 100)

    def test_progress(self):
        import sys
        import os
        import atooms.core
        realstdout = sys.stdout
        fh = open(os.devnull, 'w')
        sys.stdout = fh
        atooms.core.progress.active = True
        s = Simulation(DryRun())
        s.run(100)
        s = Simulation(DryRun())
        s.add(lambda sim: sim, 50)
        s.run(100)
        atooms.core.progress.active = False
        sys.stdout = realstdout

    def tearDown(self):
        rmd('/tmp/test_simulation')


if __name__ == '__main__':
    unittest.main()



# This file is part of atooms
# Copyright 2010-2024, Daniele Coslovich

"""Simulation backend for RUMD (http://rumd.org/)."""

# This enables importing the top-level rumd package and still call
# this file rumd.py
from __future__ import absolute_import

import sys
import os
import numpy
import logging

import rumd  # pylint:disable=import-error
from rumd.Simulation import Simulation  # pylint:disable=import-error
from atooms.system import System as _System
from atooms.system import Particle, Cell

_log = logging.getLogger(__name__)
_version = rumd.GetVersion()

if int(_version.split('.')[0]) < 3:
    raise ImportError('RUMD version must be at least 3')


def unfold(system):
    # s = system
    # particle = system.particle
    # for i, p in enumerate(particle):
    #     p.position += p.periodic_image * s.cell.side
    # system.particle = particle
    # return s
    npart = system.sample.GetNumberOfParticles()
    pos = system.sample.GetPositions()
    nsp = system.sample.GetNumberOfTypes()
    ima = system.sample.GetImages()
    spe = numpy.ndarray(npart, dtype=int)
    ii = 0
    for i in range(nsp):
        ni = system.sample.GetNumberThisType(i)
        spe[ii: ii + ni] = i
        ii += ni
    particle = [Particle(species=spe_i, position=pos_i) for spe_i, pos_i in
                zip(spe, pos)]
    for p, i in zip(particle, ima):
        p.position += i * system.cell.side

    return _System(particle=particle, cell=system.cell)


class RUMD(object):

    """RUMD simulation backend."""

    version = _version

    def __init__(self, input_file_or_sim_or_system, potentials=None,
                 integrator=None, temperature=None, dt=0.001,
                 fixcm_interval=0, thermostat_relaxation_time=None):

        # Keep a reference of the Trajectory backend class
        self.trajectory_class = Trajectory
        self.timestep = dt

        # If we pass a System instance, we write a temporary RUMD
        # trajectory to initialize the simulation and reset the
        # input_file_or_sim_or_system variable
        if isinstance(input_file_or_sim_or_system, _System):
            import tempfile
            import os
            from atooms.trajectory import TrajectoryRUMD
            tmp = os.path.join(tempfile.mkdtemp(), 'config.xyz')
            with TrajectoryRUMD(tmp, 'w') as th:
                th.write(input_file_or_sim_or_system)
            input_file_or_sim_or_system = tmp

        # Store internal rumd simulation instance.
        # It is exposed as self.rumd_simulation for further customization
        if isinstance(input_file_or_sim_or_system, Simulation):
            self.rumd_simulation = input_file_or_sim_or_system
            self._suppress_all_output = False
            self._initialize_output = True

        else:
            self.rumd_simulation = Simulation(input_file_or_sim_or_system, verbose=False)
            self.rumd_simulation.SetVerbose(False)
            self.rumd_simulation.sample.SetVerbose(False)
            self.rumd_simulation.sample.EnableBackup(False)
            self.rumd_simulation.SetMomentumResetInterval(fixcm_interval)
            self.rumd_simulation.SetBlockSize(sys.maxsize)
            self.rumd_simulation.write_timing_info = False

            # By default we mute RUMD output.
            self.rumd_simulation.SetOutputScheduling("energies", "none")
            self.rumd_simulation.SetOutputScheduling("trajectory", "none")
            self._suppress_all_output = True
            self._initialize_output = False

        # Set the forcefield
        if potentials is not None:
            # We are provided a list of rumd potentials
            for potential in potentials:
                self.rumd_simulation.AddPotential(potential)

        # Add a rumd integrator
        if temperature is not None:
            integrator = 'nvt'

        if integrator is not None:
            if integrator in ['nvt', 'NVT']:
                itg = rumd.IntegratorNVT(targetTemperature=temperature,
                                         timeStep=dt)
                if thermostat_relaxation_time is not None:
                    itg.SetRelaxationTime(float(thermostat_relaxation_time))

            elif integrator in ['nve', 'NVE']:
                itg = rumd.IntegratorNVE(timeStep=dt)
            self.rumd_simulation.SetIntegrator(itg)

        # Copy of initial state
        self._initial_sample = self.rumd_simulation.sample.Copy()

        # Hold a reference to the system
        # self.system = System(self.rumd_simulation.sample, self.rumd_simulation.potentialList)

    # Wrapping system is needed because rumd holds a reference to the
    # potentials in rumd_simulation and they are needed to create a
    # working sample from scratch
    #
    # It appears mostly necessary because we must operate on the underlying sample.
    #
    def _get_system(self):
        system = System(self.rumd_simulation.sample)
        # system.__itg_infoStr_start = self.rumd_simulation.itg.GetInfoString(8)
        return system

    def _set_system(self, value):
        self.rumd_simulation.sample.Assign(value.sample)
        # TODO: improve copying over of thermostat
        # self.rumd_simulation.itg.InitializeFromInfoString(value._itg_infoStr_start)
        # Setting sample this way is useless.
        #   self.rumd_simulation.sample = value.sample
        # Rumd actually sets samples via a file, there seems to be no other way.
        # TODO: to retain modifications to system, use atooms trajectory but at the moment we would loose info on thermostat
        # import tempfile
        # from atooms.core.utils import rmd
        # # Why should we set the output dir? It should not change
        # #tmp = value.sample.GetOutputDirectory()
        # dirout = tempfile.mkdtemp()
        # file_tmp = os.path.join(dirout, 'sample.xyz.gz')
        # value.sample.WriteConf(file_tmp)
        # self.rumd_simulation.sample.ReadConf(file_tmp, init_itg=False)
        # # Why should we set the output dir? It should not change
        # # value.sample.SetOutputDirectory(tmp)
        # # Clean up
        # rmd(dirout)

    system = property(_get_system, _set_system, 'System')

    def __str__(self):
        return 'RUMD'

    @property
    def rmsd(self):
        """
        Compute the mean square displacement between actual sample and the
        reference sample.
        """
        if self.rumd_simulation.sample is self._initial_sample:
            raise Exception('rmsd between two references of the same system does not make sense (use deepecopy?)')
        ndim = 3  # hard coded
        N = self.rumd_simulation.sample.GetNumberOfParticles()
        L = [self.rumd_simulation.sample.GetSimulationBox().GetLength(i) for i in range(ndim)]
        # Unfold positions using periodic image information
        ref = self._initial_sample.GetPositions() + self._initial_sample.GetImages() * L
        unf = self.rumd_simulation.sample.GetPositions() + self.rumd_simulation.sample.GetImages() * L
        return (sum(sum((unf - ref)**2)) / N)**0.5

    def write_checkpoint(self, output_path):
        with Trajectory(output_path + '.chk', 'w') as t:
            t.write(self.system, None)

    def read_checkpoint(self, output_path):
        if os.path.exists(output_path + '.chk'):
            self.rumd_simulation.sample.ReadConf(output_path + '.chk')
        else:
            _log.debug('could not find checkpoint')

    def run(self, steps):
        self.rumd_simulation.Run(steps,
                                 suppressAllOutput=self._suppress_all_output,
                                 initializeOutput=self._initialize_output)
        self._initialize_output = False

class Thermostat(object):

    """Wrap a RUMD integrator as a thermostat."""

    # TODO: problem, RUMD must keep the same order in future versions
    # We should unit test this
    # Info string looks like IntegratorNVT,0.004,0.3602,0.2,-0.7223

    def __init__(self, integrator):
        self._integrator = integrator

    def reset(self):
        info = self._integrator.GetInfoString(18).split(',')
        info[4] = '1.0'
        info = ','.join(info)
        self._integrator.InitializeFromInfoString(info)

    def _get_temperature(self):
        info = self._integrator.GetInfoString(18).split(',')
        return float(info[2])

    def _set_temperature(self, value):
        info = self._integrator.GetInfoString(18).split(',')
        info[2] = '%g' % value
        info = ','.join(info)
        self._integrator.InitializeFromInfoString(info)

    temperature = property(_get_temperature, _set_temperature, 'Temperature')


class System(object):

    """System wrapper for RUMD."""

    def __init__(self, sample):
        self.sample = sample
        self.thermostat = Thermostat(self.sample.GetIntegrator())
        self.barostat = None
        self.reservoir = None

    def __copy__(self):
        # This is not really needed, it's just there for reference
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        result.sample = self.sample.Copy()
        return result

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        # result.__dict__.update(self.__dict__)  #these are refs...
        # Copy() does not copy the integrator
        result.sample = self.sample.Copy()
        # This way we set the integrator. Note that it is always the same object...
        # result.sample.SetIntegrator(self.sample.GetIntegrator())
        # result._itg_infoStr_start = self.thermostat._integrator.GetInfoString(18)
        # TODO: thermostat should be a property this way we would not need to update this
        result.thermostat = Thermostat(self.sample.GetIntegrator())
        return result

    def update(self, other):
        self.sample.Assign(other.sample)
        # self.sample.SetIntegrator(other.sample.GetIntegrator())  # maybe not necessary
        self.thermostat = Thermostat(self.sample.GetIntegrator())

    def potential_energy(self, per_particle=False, normed=False, cache=False):
        self.sample.CalcF()
        if normed or per_particle:
            return self.sample.GetPotentialEnergy() / len(self.particle)
        else:
            return self.sample.GetPotentialEnergy()

    def kinetic_energy(self, per_particle=False, normed=False):
        # TODO: use double IntegratorNVT::GetKineticEnergy(bool copy) const{
        ekin = sum([p.kinetic_energy for p in self.particle])
        if normed or per_particle:
            return ekin / len(self.particle)
        else:
            return ekin

    def total_energy(self, per_particle=False, normed=False, cache=False):
        return self.potential_energy(per_particle=per_particle, normed=normed, cache=cache) +\
            self.kinetic_energy(per_particle=per_particle, normed=normed)

    def __get_mass(self):
        # TODO: cache it (but what if masses change?)
        npart = self.sample.GetNumberOfParticles()
        nsp = self.sample.GetNumberOfTypes()
        mass = numpy.ndarray(npart, dtype=float)
        ii = 0
        for i in range(nsp):
            ni = self.sample.GetNumberThisType(i)
            try:
                # This will work with rumd <= 2.0.1 I think
                # meta = self.sample.GetTrajectoryConfMetaData()
                # then get meta.GetMassOfType(i)
                mi = self.sample.GetMass(i)
            except:
                _log.warn('cannot get mass from RUMD interface, setting to 1.0')
                mi = 1.0
            mass[ii: ii + ni] = mi
            ii += ni
        return mass

    @property
    def temperature(self):
        ndof = self.sample.GetNumberOfDOFs()
        vel = self.sample.GetVelocities()
        mass = self.__get_mass()
        return numpy.sum(mass * numpy.sum(vel**2.0, 1)) / ndof

    # TODO: this should be the attribute setter
    def set_temperature(self, T):
        # Scale velocities from temperature Told to T
        # TODO: use maxwellian
        # TODO: subtract CM velocity
        Told = self.temperature
        velocity_factor = (T/Told)**0.5
        self.sample.ScaleVelocities(velocity_factor)

    def scale_velocities(self, factor):
        self.sample.ScaleVelocities(factor)

    @property
    def cell(self):
        box = self.sample.GetSimulationBox()
        L = [box.GetLength(i) for i in range(3)]
        return Cell(L)

    @property
    def particle(self):
        # Warning n.1 : this is read only. If you change the particles, the
        # modification won't be propoagated to the RUMD objects.
        # One would have to create a new system.
        #
        # Warning n.2 : it ia assumed that particles are sorted by species.
        # since RUMD does not have accessors to particle types (why??)
        # and we can only access the number of particles of a given type.
        npart = self.sample.GetNumberOfParticles()
        pos = self.sample.GetPositions()
        vel = self.sample.GetVelocities()
        nsp = self.sample.GetNumberOfTypes()
        ima = self.sample.GetImages()
        mass = self.__get_mass()
        spe = numpy.ndarray(npart, dtype=int)
        ii = 0
        for i in range(nsp):
            ni = self.sample.GetNumberThisType(i)
            spe[ii: ii + ni] = i
            ii += ni
        p = [Particle(species=spe_i, mass=mass_i, position=pos_i,
                      velocity=vel_i) for spe_i, mass_i, pos_i, vel_i
             in zip(spe, mass, pos, vel)]
        for pi, i in zip(p, ima):
            pi.periodic_image = i
        return p

    def dump(self, what):
        import atooms.system
        system = atooms.system.System(self.particle, self.cell)
        return system.dump(what)


class Trajectory(object):

    suffix = 'xyz'

    def __init__(self, filename, mode='w'):
        self.filename = filename
        self.mode = mode

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def write(self, system, step):
        """
        If step is not None, output will follow a folder-based logic and
        filename will be considered as the root folder
        """
        if step is None:
            f = self.filename
        else:
            fbase = '%011d.%s' % (step, self.suffix)
            f = os.path.join(self.filename, fbase)
            if not os.path.exists(self.filename):
                os.makedirs(self.filename)
        _log.debug('writing config via backend to %s at step %s, %s', f, step, self.mode)
        system.sample.WriteConf(f, self.mode)

    def close(self):
        # This only unzips files with no step info
        if os.path.exists(self.filename + '.gz'):
            os.system("gunzip -f %s.gz" % self.filename)



# This file is part of atooms
# Copyright 2010-2024, Daniele Coslovich

"""
Trajectory callbacks and class decorators.

- "Callbacks" are simple functions that modify the `System` instance
  returned by `read_system()`. They can be registered to `trajectory`
  instance via `add_callback()`.

- "Class decorators" can be used for more complex modifications of
  trajectory behavior. They return dynamically subclassed trajectory
  instance.
"""

import numpy
import copy
from atooms.system import System, Molecule

__all__ = ['center', 'change_species', 'sort', 'filter_species',
           'set_density', 'set_temperature', 'fix_cm', 'fold',
           'create_molecular_system', 'Sliced', 'Unfolded']


# Callbacks

def center(system):
    """
    Shift the origin of reference frame at the center of the
    simulation cell. Particles positions will thus lie between -L/2
    and L/2, where L is the cell side along each direction.

    This function does not check if that is done multiple times.
    """
    system.cell.center[:] = 0
    for p in system.particle:
        p.position -= system.cell.side / 2.0
    return system

def change_species(system, layout):
    """
    Return a modified `system` with particle species changed according
    to `layout`.

    The possible values of `layout` are:

    - 'A': alphabetic, i.e. species are strings like 'A', 'B', ...
    - 'C': C-like, i.e. species are integers starting from 0
    - 'F': F-like, i.e. species are integers starting from 1

    If the current layout already matches the requested one, the
    system is returned unchanged.
    """
    if layout not in ['A', 'C', 'F']:
        raise ValueError('species layout must be A, C, or F (not %s)' % layout)

    # Detect species layout (A=alphabetic, C=C style, F=fortran style)
    try:
        species = [int(p.species) for p in system.particle]
    except ValueError:
        # The species cannot be converted to int, thus layout is
        # alphabetical
        current_layout = 'A'
    else:
        min_sp = numpy.min(species)
        if min_sp == 0:
            current_layout = 'C'
        elif min_sp == 1:
            current_layout = 'F'
        else:
            raise ValueError('Numeric species should start from 0 or 1')

    # Do nothing if the layout is already ok
    if layout == current_layout:
        return system

    # Convert to new layout
    import string
    if layout == 'A':
        # We get the index of the species map:
        # - if current layout is F (min_sp=1), we subtract one.
        # - if current layout is C (min_sp=0), we do nothing
        species_map = string.ascii_uppercase
        for p in system.particle:
            p.species = species_map[int(p.species) - min_sp]
    else:
        # Output layout is numerical (C or F)
        from atooms.system.particle import distinct_species
        offset = 1 if layout == 'F' else 0
        # Note that distinct_species is sorted alphabetically
        species_list = distinct_species(system.particle)
        if current_layout == 'A':
            for p in system.particle:
                p.species = str(species_list.index(p.species) + offset)
        else:
            # If layout=C, current_layout is F and we subtract 2*offset-1=-1
            # If layout=F, current_layout is C and we add 2*offset-1=+1
            for p in system.particle:
                p.species = str(int(p.species) + 2*offset - 1)
    return system

def sort(system):
    """Sort particles by species id."""
    system.particle = sorted(system.particle, key=lambda a: a.species)
    return system

def filter_species(system, species):
    """Return particles of a given `species` id."""
    system.particle = [p for p in system.particle if p.species == species]
    return system

def set_density(system, rho):
    """Set density of system to `rho` by rescaling the cell."""
    rho_old = system.density
    x = (rho_old / rho)**(1./3)
    system.cell.side *= x
    for p in system.particle:
        p.position *= x
    return system

def set_temperature(system, T):
    """Set system temperature to `T` by reassigning velocities."""
    from atooms.system.particle import cm_velocity
    for p in system.particle:
        p.maxwellian(T)
    v_cm = cm_velocity(system.particle)
    for p in system.particle:
        p.velocity -= v_cm
    return system

def fix_cm(s):
    """
    Return a system whose positions are relative to the original
    system's center of mass.
    """
    # Get current position of CM from unfolded positions
    cm = s.cm_position
    for p in s.particle:
        p.position -= cm
    return s

def fold(s):
    """Center and fold positions into central cell."""
    for p in s.particle:
        p.position -= s.cell.side / 2
        p.fold(s.cell)
    return s

def create_molecular_system(particle_system):
    """
    Convert a particle system to a molecular system.

    Args:
        particle_system (System): The particle system to convert.

    Returns:
        System: The molecular system.
    """
    molecules, ps = {}, []
    for p in particle_system.particle:
        if not hasattr(p, 'molecule_id'):
            ps.append(p)
            continue
        molecule_id = p.molecule_id
        if molecule_id not in molecules:
            molecules[molecule_id] = []
        molecules[molecule_id].append(p)
    if len(molecules) == 0:
        raise AttributeError("No molecules found (missing 'molecule_id' attribute). Cannot create molecular trajectory.")

    ms = [Molecule(ps, bond=[], cell=particle_system.cell) for ps in molecules.values()]
    return System(particle=ps, molecule=ms, cell=particle_system.cell)


# Class decorators

# Some of these decorators raise pylint errors because of the
# lack of visibility on parent methods() or attributes

# To properly implement decorators in python see
# http://stackoverflow.com/questions/3118929/implementing-the-decorator-pattern-in-python
# asnwer by Alec Thomas. if we don't subclass at runtime we won't be able to use the decorated
# mathod in other non-subclassed methods.

class Sliced(object):

    """Only return a slice of a trajectory."""

    # This is still necessary. slicing via __getitem__ has a large memory fingerprint
    # since we couldnt write it as a generator (maybe it is possible?)
    # TODO: adjust uslice to pick up blocks without truncating them

    def __new__(cls, component, uslice):
        import copy
        cls = type('Sliced', (Sliced, component.__class__), component.__dict__)
        return object.__new__(cls)

    def __init__(self, component, uslice):
        # pylint:disable=access-member-before-definition
        self._sliced_frames = range(len(self.steps))[uslice]
        self.steps = self.steps[uslice]
        # Reset cache (this fix was long due...)
        self._cache = None
        self._initialized_read = False
        # Ensure callbacks are not shared from this moment onwards
        import copy
        self.callbacks = copy.copy(component.callbacks)

    def read_system(self, frame):
        i = self._sliced_frames[frame]
        return super(Sliced, self).read_system(i)


class Unfolded(object):

    """Decorate Trajectory to unfold particles positions on the fly."""

    def __new__(cls, component, fixed_cm=False):
        cls = type('Unfolded', (Unfolded, component.__class__), component.__dict__)
        return object.__new__(cls)

    def __init__(self, component, fixed_cm=False):
        self._component = component
        self._cache = None  # reset cache
        self._initialized_read = False
        # Ensure callbacks are not shared from this moment onwards
        import copy
        self.callbacks = copy.copy(component.callbacks)
        self.fixed_cm = fixed_cm

    def read_init(self):
        s = super(Unfolded, self).read_init()
        # Cache the initial sample and cell
        s = copy.deepcopy(self._component.read(0))
        self._old = numpy.array([p.position for p in s.particle])
        self._last_read = 0

    def read_system(self, frame):
        # Return here if first frame
        if frame == 0:
            # Deepcopy needed, see below
            s = copy.deepcopy(self._component.read(frame))
            if self.fixed_cm:
                s = fix_cm(s)
            return s

        # Compare requested frame with last read
        delta = frame - self._last_read
        if delta < 0:
            raise ValueError('cannot unfold jumping backwards (delta=%d)' % delta)
        if delta > 1:
            # Allow to skip some frames by reading them internally
            # We read delta-1 frames, then delta is 1
            for _ in range(delta-1):
                self.read_system(self._last_read+1)

        # With deepcopy we make sure that Unfolded() returns copies of
        # the system read by the component trajectory and does not
        # modify the underlying system, which is important in case the
        # latter is stored in memory (TrajectoryRam) and with caching
        s = copy.deepcopy(self._component.read(frame))
        self._last_read = frame

        # Unfold positions
        # Note that since L can be variable we get it at each step
        # TODO: I am not entirely sure this is correct with NPT.
        # The best thing in this case is to get unfolded positions
        # from the simulation.
        L = s.cell.side
        pos = numpy.array([p.position.copy() for p in s.particle])
        dif = pos - self._old
        dif = dif - numpy.rint(dif / L) * L
        self._old += dif

        # Copy unfolded positions back to the system
        # Here we cannot do
        #   s.particle[i].position = self._old[i][:]
        # because this a shallow view and the arrays share memory.
        # Fixing the CM later on will not work correctly.
        for i in range(len(pos)):
            s.particle[i].position = self._old[i].copy()

        if self.fixed_cm:
            s = fix_cm(s)

        return s


# Not necessary for the time being
# class _Molecular:
#     """
#     Create a molecular trajectory from a particle trajectory.
#     """

#     def __new__(cls, component):
#         cls = type('Molecular', (Molecular, component.__class__), component.__dict__)
#         return object.__new__(cls)

#     def __init__(self, particle_trajectory):
#         """
#         Initialize the Molecular trajectory with a particle trajectory.

#         Args:
#             particle_trajectory (TrajectoryBase): The particle trajectory to convert.
#         """
#         # pylint:disable=access-member-before-definition
#         self._particle_trajectory = particle_trajectory

#     def read_system(self, frame):
#         """
#         Read the system for a given frame and convert it to a molecular system.

#         Args:
#             frame (int): The frame index to read.

#         Returns:
#             System: The molecular system for the given frame.
#         """
#         particle_system = self._particle_trajectory.read_system(frame)
#         return create_molecular_system(particle_system)


def _Molecular(cls):

    class Molecular(cls):

        def read_system(self, frame):
            """
            Read the system for a given frame and convert it to a molecular system.

            Args:
                frame (int): The frame index to read.

            Returns:
                System: The molecular system for the given frame.
            """
            particle_system = super().read_system(frame)
            return create_molecular_system(particle_system)

        def write_system(self, system, step):
            """
            Write the system for a given frame, adding the molecule_id variable

            Args:
                system (System): the system instance to be written
                step (int): The step index being written.
            """
            if 'particle.molecule_id' not in self.variables:
                self.variables.append('particle.molecule_id')
            for i, molecule in enumerate(system.molecule):
                for p in molecule.particle:
                    p.molecule_id = i
            super().write_system(system, step)

    return Molecular



# This file is part of atooms
# Copyright 2010-2024, Daniele Coslovich

"""
Base trajectory classes.
"""

import os
import copy
import warnings


from .utils import get_block_size


# FutureWarning must show a traceback so that the user can localize the deprecation
# This should be moved up the chain
def _warn_with_traceback(message, category, filename, lineno, file=None, line=None):
    import traceback
    import sys
    if category is FutureWarning:
        traceback.print_stack(file=sys.stderr, limit=6)
    sys.stderr.write(warnings.formatwarning(message, category, filename, lineno, line))


warnings.showwarning = _warn_with_traceback


def canonicalize_fields(fields):
    from atooms.core.utils import canonicalize

    warnings.warn('canonicalize_fields() is deprecated, use canonicalize() instead', FutureWarning)
    th = TrajectoryBase(None)
    return canonicalize(fields, th.thesaurus)


class TrajectoryBase(object):
    """
    Trajectory abstract base class.

    A trajectory is composed by one or several frames, each frame
    being a sample of a `System` taken at a given `step` during a
    simulation. Trajectory instances are iterable and behave as file
    objects: they can be opened and closed using the `with` syntax

    .. code-block:: python

       with Trajectory(inpfile) as th:
           for system in th:
               pass

    alternatively

    .. code-block:: python

       th = Trajectory(inpfile)
       for system in th:
           pass
       th.close()

    To write the state of a `System` to a trajectory, we must open the
    trajectory in write mode

    .. code-block:: python

       with Trajectory(outfile, 'w') as th:
           th.write(system, step=0)

    Trajectories can use the `variables` attribute to define which system
    properties will be written to disk. Concrete classes may thus
    provide means to write arbitrary variables to disk via particle
    properties. 

    Examples
    --------

    .. code-block:: python

       for particle in system:
           particle.some_custom_property = 1.0
       with Trajectory(outfile, 'w') as th:
           th.variables = ['position', 'some_custom_property']
           th.write(system, step=0)

    To be fully functional, concrete classes must implement
    `read_system()` and `write_system()` methods.

    The `read()` method is a template composed of the two following steps:

    - `read_init()`: called only once to initialize internal data structures,
      grab metadata, etc. Need *not* be implemented by subclasses.

    - `read_system()`: actually return the system at a given frame. It must
      be implemented by subclasses.

    Similarly, `write()` is a template composed of `write_init()` and
    `write_system()`. Only the latter method must be implemented by
    subclasses.

    The `cache` variable reduces access time when reading the same
    trajectory multiple times. We use shallow copies to cut down the
    overhead. Cache is disabled by default as there is no control on
    its size yet.
    """

    suffix = None

    def __init__(self, filename, mode='r', cache=False):
        self.filename = filename
        self.mode = mode
        """Can be 'r' (read) or 'w' (write)."""
        self.callbacks = []
        """List of callbacks to be applied to the system when reading a frame, see `add_callback()`"""
        self.precision = 6
        """Number of digits used to write trajectory files."""
        self.metadata = {}
        """
        Dictionary of metadata about the trajectory. It can be used by
        subclasses to hold trajectory format info or dynamically
        on a per sample basis.
        """
        self.thesaurus = {
            'position': 'particle.position',
            'velocity': 'particle.velocity',
            'species': 'particle.species',
            'radius': 'particle.radius',
            'mass': 'particle.mass',
            'pos': 'particle.position',
            'x': 'particle.position[0]',
            'y': 'particle.position[1]',
            'z': 'particle.position[2]',
            'vel': 'particle.velocity',
            'vx': 'particle.velocity[0]',
            'vy': 'particle.velocity[1]',
            'vz': 'particle.velocity[2]',
            'id': 'particle.species',
            'type': 'particle.species',
            'name': 'particle.species',
        }
        """
        Dictionary of common shortcuts and synonims for system
        attributes. Can be updated by subclasses.
        """

        # These are cached properties
        self._variables = ()
        """
        Tuple of system attributes to be written by `write_system` and/or
        read by `read_system`. Its entries are canonicalized using
        `self.thesaurus` everytime the attribute is set. Subclasses
        may use it to allow the user to modify the trajectory layout
        or they can ignore it entirely.
        """
        self._steps = None
        self._timestep = None
        self._grandcanonical = None
        self._block_size = None

        # Internal state
        self._overwrite = False
        self._initialized_write = False
        self._initialized_read = False
        # Sanity checks
        # Subclasses may define mode variants, such as r+, therefore the
        # autoritative mode is the first letter
        if self.mode[0] == 'r' and self.filename is not None and \
           not os.path.exists(self.filename):
            raise IOError('trajectory file %s does not exist' % self.filename)
        # Cache frames to optimize reading the same trajectory multiple times
        # We use shallow copies to cut down the overhead
        self.cache = cache
        """If `True`, use a cache when reading the same frame multiple times"""
        self._cache = None

    def __len__(self):
        # We try first with read_len() which returns None by default
        frames = self.read_len()
        if frames is None:
            # We get the steps, which might take a bit longer
            return len(self.steps)
        else:
            return frames

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def __iter__(self):
        for i in range(len(self.steps)):
            yield self.read(i)

    def __getitem__(self, key):
        if isinstance(key, slice):
            # This works but it loads the whole trajectory in ram.
            # If we make this a generator, then access a single sample
            # wont work.
            # The Sliced decorator doesn't have this issue.
            # Or just slice range(self.steps) instead.
            frames = range(len(self.steps))
            return [self.read(i) for i in frames[key]]

        elif isinstance(key, int):
            if key < 0:
                key += len(self)
            if key >= len(self):
                raise IndexError("Index (%d) is out of range (%d)." % (key, len(self)))
            return self.read(key)

        else:
            raise TypeError("Invalid argument type [%s]" % type(key))

    def append(self, system):
        """Equivalent to `write(system)`."""
        self.write(system)

    def close(self):
        """Can be implemented by subclasses to close file handles."""
        pass

    def read(self, index):
        """Read and return system at frame `index`."""
        if not self._initialized_read:
            self.read_init()
            self._initialized_read = True

        if self.cache and self._cache and index in self._cache:
            # We get the system from the cache
            system = self._cache[index]
        else:
            system = self.read_system(index)
            if self.cache:
                # Store the system in cache
                if self._cache is None:
                    self._cache = {}
                self._cache[index] = copy.copy(system)

        # Apply callbacks, class level and instance level
        #
        # Pass the frame index to the callback via system
        # by storing a "temporary" frame instance variable
        # (deleted after this loop)
        system.frame = index
        # Copying callbacks is an old hack that allowed client
        # code to safely pass variables in the callback across calls
        # Deprecating this feature would require checking if the
        # callback has variables not starting with __
        callbacks = copy.copy(self.callbacks)
        for cbk, args, kwargs in callbacks:
            system = cbk(system, *args, **kwargs)
        if hasattr(system, 'frame'):
            del (system.frame)
        return system

    def write(self, system, step=None):
        """
        Write `system` at a given integer `step`.

        If `step` is not provided, it is defined internally by
        incrementing by one the last added step, staring from zero.
        """
        if self.mode[0] == 'r':
            raise IOError('trajectory file not open for writing')
        if not self._initialized_write:
            self.write_init(system)
            self._initialized_write = True

        # If we do not provide a step, we incrementally add 1 to the
        # last step, starting from 0.
        if step is not None:
            current_step = step
        else:
            if len(self.steps) == 0:
                current_step = 0
            else:
                current_step = self.steps[-1] + 1

        # If overwriting is not allowed (default), we check that we
        # are adding a step larger than the last added step.
        if not self._overwrite:
            if len(self.steps) > 0 and current_step <= self.steps[-1]:
                raise ValueError('cannot add step {} when overwrite is False'.format(current_step))

        # Write the sample.
        self.write_system(system, current_step)
        # Step is added last, frame index starts from 0 by default.
        if step is None:
            self.steps.append(current_step)
        else:
            # If overwriting is allowed, we append the step only if it
            # not already there. This enables a small optimization by
            # avoiding this check if overwriting is off.
            if not self._overwrite:
                self.steps.append(current_step)
            else:
                if current_step not in self.steps:
                    self.steps.append(current_step)

    def read_init(self):
        """
        Read metadata and/or set up data structures. Need not be
        implemented.
        """
        pass

    def write_init(self, system):
        """Subclass should use it to open files for writing."""
        pass

    # These methods must be implemented by subclasses

    def read_system(self, frame):
        """Return the system at the given `frame`."""
        if hasattr(self, 'read_sample'):
            warnings.warn('read_sample() is deprecated, use read_system()', DeprecationWarning)
            return self.read_sample(frame)
        else:
            raise NotImplementedError()

    def write_system(self, system, step):
        """Write a `system` to file. Noting to return."""
        if hasattr(self, 'write_sample'):
            warnings.warn('write_sample() is deprecated, use write_system()', DeprecationWarning)
            return self.write_sample(system, step)
        else:
            raise NotImplementedError()

    # Callbacks will be applied to the output of read_system()

    def register_callback(self, cbk, *args, **kwargs):
        """
        Register a callback `cbk` to be applied when reading a frame.

        The callback must receive a `System` instance as input an
        return the modified `System` instance as output.
        """
        if (cbk, args, kwargs) not in self.callbacks:
            self.callbacks.append((cbk, args, kwargs))

    def add_callback(self, cbk, *args, **kwargs):
        """Same as `register_callback()`."""
        self.register_callback(cbk, *args, **kwargs)

    # To read/write timestep and block size sublcasses may implement
    # these methods. The default is dt=1 and block determined dynamically.

    def read_len(self):
        """Return the number of frames. Optional."""
        return None

    def read_steps(self):
        """Return a list of steps."""
        return []

    def read_timestep(self):
        """Can be subclassed to parse the timestep."""
        return 1.0

    def write_timestep(self, value):
        """Set the timestep."""
        self._timestep = value

    def read_block_size(self):
        """
        Can be subclassed to parse the block size for trajectory with
        non-uniform intervals between frames (ex. exponential sampling).
        """
        return None

    def write_block_size(self, value):
        """Can be subclassed to write the block size."""
        pass

    # Properties

    @property
    def variables(self):
        return self._variables

    @variables.setter
    def variables(self, value):
        from atooms.core.utils import canonicalize
        self._variables = canonicalize(value, self.thesaurus)

    def copy(self, cls=None, fout=None, only=None, include=None, exclude=None, steps=None):
        """
        Return a copy of the trajectory
        """
        from atooms.core.progress import progress
        from atooms.core.utils import canonicalize
        from atooms.core.utils import mkdir

        # Output class
        if cls is None:
            cls = self.__class__
        else:
            from atooms.trajectory import Trajectory
            if isinstance(cls, str):
                cls = Trajectory.formats[cls]

        # Make sure parent folder exists
        if fout is not None:
            mkdir(os.path.dirname(fout))

        # Copy trajectory
        conv = cls(fout, 'w')
        # In a previous version of trajectory conversion, we only
        # included the variables of the original trajectory, to allow
        # the output trajectory to include variables not present in
        # the original one. I do not think this makes sense. Suppose
        # the original trajectory did not store velocities, but the
        # output one writes them by default. The results will be wrong.
        variables = list(self.variables)
        if only is not None:
            variables = only
        if include is not None:
            for pattern in canonicalize(include, self.thesaurus):
                if pattern not in variables:
                    variables.append(pattern)
        if exclude is not None:
            for pattern in canonicalize(exclude, self.thesaurus):
                if pattern not in variables:
                    variables.append(pattern)
        conv.variables = variables
        conv.precision = self.precision
        conv.timestep = self.timestep
        conv.block_size = self.block_size
        # In python 3, zip returns a generator so this is ok
        #
        # for system, step in zip(inp, inp.steps):
        #     conv.write(system, step)
        #
        # In python 2, zipping t and t.steps will load everything
        # in RAM. In this case, it is better to use enumerate()
        if steps is None:
            for i, system in progress(enumerate(self), total=len(self)):
                conv.write(system, self.steps[i])
        else:
            # Only include requested steps (useful to prune
            # non-periodic trajectories)
            for step in steps:
                idx = self.steps.index(step)
                conv.write(self[idx], step)
        return conv

    @property
    def fields(self):
        warnings.warn('fields is deprecated, use variables instead', FutureWarning)
        return self.variables

    @fields.setter
    def fields(self, value):
        warnings.warn('fields is deprecated, use variables instead', FutureWarning)
        self._variables = value

    @property
    def steps(self):
        if self._steps is None:
            if self.mode[0] == 'r':
                self._steps = self.read_steps()
            else:
                self._steps = []
        return self._steps

    @steps.setter
    def steps(self, value):
        self._steps = value

    @property
    def timestep(self):
        if self._timestep is None:
            if self.mode[0] == 'r':
                self._timestep = self.read_timestep()
            else:
                self._timestep = 1.0
        return self._timestep

    @timestep.setter
    def timestep(self, value):
        self.write_timestep(value)
        self._timestep = value

    @property
    def block_size(self):
        """
        Return the block size in frames of the trajectory.

        A block is a sequence of steps that repeats periodically in
        the trajectory. The block size is 1 for trajectories with
        constant time intervals between frames. A block size larger
        than 1 occurs with exponential sampling, ex. for steps
        sequences like 1,2,4,11,12,14,...
        """
        if self._block_size is None:
            self._block_size = self.read_block_size()
        if self._block_size is None:
            # If size is still None (read_block_size is not
            # implemented) we determine it dynamically
            self._block_size = get_block_size(self.steps)
        return self._block_size

    @block_size.setter
    def block_size(self, value):
        self._block_size = value
        self.write_block_size(value)

    # Some additional useful properties

    @property
    def grandcanonical(self):
        """
        True if the trajectory is grandcanonical, i.e. the number of
        particles changes.
        """
        # In subclasses, cache it for efficiency, since we might have to discover it
        if self._grandcanonical is None:
            self._grandcanonical = False
        return self._grandcanonical

    @property
    def times(self):
        """All available times."""
        return [s*self.timestep for s in self.steps]

    @property
    def total_time(self):
        """Total simulation time."""
        return (self.steps[-1] - self.steps[0]) * self.timestep


class SuperTrajectory(TrajectoryBase):

    """Collection of subtrajectories."""

    # Optimized version

    def __init__(self, files, trajectoryclass, mode='r'):
        """
        Group a list of `files` into a single trajectory of class
        `trajectoryclass`.
        """
        self.files = files
        if len(self.files) == 0:
            raise ValueError('no files found in %s' % self.files)
        f = os.path.dirname(self.files[0])
        super(SuperTrajectory, self).__init__(f, mode)
        self.trajectoryclass = trajectoryclass

        # Make sure subtrajectories are sorted by increasing step
        # We look up the first step in each subtrajectory
        _step = []
        for f in self.files:
            with self.trajectoryclass(f) as t:
                _step.append(t.steps[0])
        _, self.files = zip(*sorted(zip(_step, self.files)))
        self.files = list(self.files)

        # This list holds the file containing a given step
        self._steps_file = []
        self._steps_frame = []
        # This caches the last trajectory used to minimize __init__() overhead
        self._last_trajectory = None
        self.steps = []
        for f in self.files:
            # This is slow, just to get the step index.
            # If we accept not to have the steps list updated at this stage
            # we can optimize this by about 10% on xyz files (16.12.2016)
            with self.trajectoryclass(f) as t:
                for j, step in enumerate(t.steps):
                    # The second test is to avoid storing twice the same frame
                    # if it appears at the end of a file and the beginning
                    # of the next file
                    if len(self.steps) == 0 or step != self.steps[-1]:
                        self.steps.append(step)
                        self._steps_file.append(f)
                        self._steps_frame.append(j)

        # Propagate variables
        with self.trajectoryclass(self.files[0]) as t:
            self.variables = t.variables

    def read_system(self, frame):
        # Optimization: use the last trajectory in cache (it works
        # well if frames are read sequentially)
        f = self._steps_file[frame]
        j = self._steps_frame[frame]

        # Note: if the file object has been closed in the meantime,
        # this approach will fail. In atooms 2, there was a check
        # whether the file object (trajectory attribute) was
        # closed. We will assume that decorators have no side
        # effects. If this needs to be fixed again, we can find a way.
        if self._last_trajectory is None:
            # First trajectory
            self._last_trajectory = self.trajectoryclass(f)
        elif self._last_trajectory.filename != f:
            # This is a new trajectory file
            self._last_trajectory.close()
            self._last_trajectory = self.trajectoryclass(f)
        else:
            # Already in cache, we just pass for clarity
            pass
        t = self._last_trajectory
        return t[j]

    def read_timestep(self):
        with self.trajectoryclass(self.files[0]) as t:
            return t.timestep

    def close(self):
        if self._last_trajectory is not None:
            self._last_trajectory.close()
            self._last_trajectory = None
        super(SuperTrajectory, self).close()



#!/usr/bin/env python

"""Convert trajectory file to a different format."""

from __future__ import print_function
import os
import sys
import logging
import argparse
import random
from atooms import trajectory
from atooms.core.utils import fractional_slice, add_first_last_skip, setup_logging
from atooms.trajectory.utils import check_block_size, info, formats


def main_info(args):
    """Print info on trajectories."""
    for file_inp in args.file_inp:
        if args.folder:
            th = trajectory.folder.Foldered(file_inp, cls=args.inp)
        else:
            th = trajectory.Trajectory(file_inp, fmt=args.inp)

        if args.fields:
            print(info(th, args.fields))
        else:
            print(info(th))

def main(args):
    """Convert trajectory `file_inp` to `file_out`."""
    args.file_inp = args.file_inp[0]

    if args.fields is not None:
        args.fields = args.fields.split(',')

    if args.out is not None and not args.out in trajectory.Trajectory.formats:
        formats()
        raise ValueError('Unknown output format %s' % args.out)

    if args.file_out == '-':
        args.file_out = '/dev/stdout'

    if args.verbose:
        setup_logging('atooms', 20)
        import atooms.core.progress
        if args.file_out != '/dev/stdout':
            atooms.core.progress.active = True

    if args.folder:
        t = trajectory.folder.Foldered(args.file_inp, cls=args.inp)
    else:
        t = trajectory.Trajectory(args.file_inp, fmt=args.inp)

    # If no output format is provided we use the input one
    if args.out is None:
        out_class = t.__class__
    else:
        out_class = args.out

    if args.precision is not None:
        t.precision = args.precision

    if args.flatten_steps:
        t.steps = range(1, len(t)+1)

    # Reset random number generator
    if args.seed:
        random.seed(args.seed)

    # Trick to allow some trajectory formats to set the box side.
    # This way the cell is defined as we read the sample (callbacks
    # will not do that).
    if args.side is not None:
        def fix_cell(system, side):
            from atooms.system import Cell
            system.cell = Cell(side)
            return system
        L = args.side.split(',')
        if len(L) == 1:
            t.add_callback(fix_cell, [L, L, L])
        else:
            t.add_callback(fix_cell, [float(_) for _ in L])

    # Define slice.
    # We interpret --first N --last N as a request of step N
    if args.last == args.first and args.last is not None:
        args.last += 1
    sl = fractional_slice(args.first, args.last, args.skip, len(t))

    # Unfold if requested
    if args.unfold:
        tu = trajectory.Unfolded(t) #, fix_cm=True)
    else:
        tu = t

    # Fix CM and fold back
    if args.fix_cm:
        tu.add_callback(trajectory.fix_cm)
        tu.add_callback(trajectory.fold)

    # Here we could you a trajectory slice t[sl] but this will load
    # everything in ram (getitem doesnt provide a generator). This
    # will be fixed with python 3.
    ts = trajectory.Sliced(tu, sl)

    # Change number of particles
    if args.N > 0:
        def decimate_system(system, N):
            from atooms.system.particle import decimate
            system.particle = decimate(system.particle, N)
            return system
        ts.register_callback(decimate_system, args.N)
    # Change density and temperature
    if args.rho is not None:
        ts.register_callback(trajectory.decorators.set_density, args.rho)
    if args.temperature is not None:
        ts.register_callback(trajectory.decorators.set_temperature, args.temperature)
    # Change species layout if requested
    if args.species_layout is not None:
        ts.register_callback(trajectory.decorators.change_species, args.species_layout)
    # Sort by species id
    if args.species_sort:
        ts.register_callback(trajectory.decorators.sort)

    # We enforce regular periodicity; steps is None is trajectory is not periodic
    try:
        steps = check_block_size(ts.steps, ts.block_size, prune=True)
    except IndexError:
        print('# Warning: something wrong with periodicity check.')
        print('# We will proceed, but you should check the converted trajectory.')
        steps = ts.steps

    #
    # ---------------------
    # Trajectory conversion
    # ---------------------
    #
    include_list, exclude_list = [], []
    if len(args.fields_include) > 0:
        include_list = args.fields_include.split(',')
    if len(args.fields_exclude) > 0:
        exclude_list = args.fields_exclude.split(',')

    from atooms.trajectory import Trajectory
    if isinstance(out_class, str):
        out_class = Trajectory.formats[out_class]

    conv = ts.copy(fout=args.file_out, cls=out_class,
                   only=args.fields, exclude=exclude_list,
                   include=include_list, steps=steps)
    fout = conv.close()

    if args.ff:
        from atooms.trajectory.hdf5 import add_interaction_hdf5
        add_interaction_hdf5(fout, args.ff)

    if args.verbose and args.file_out != '/dev/stdout':
        print('# converted %s to %s' % (args.file_inp, fout))

    t.close()


def main_paste(args):
    """
    Correlate particles properties from trajectory files.

    Example:
    --------

    trj.py paste.py file1.xyz:radius file2.xyz.voronoi.xyz:volume
    """
    from atooms import trajectory as trj
    from atooms.core.utils import tipify

    f1, attr1 = args.file_inp[0].split(':')
    f2, attr2 = args.file_inp[1].split(':')
    if args.inp is None:
        fmt1, fmt2 = None, None
    else:
        fmt1, fmt2 = args.inp.split(',')
    t1 = trj.Trajectory(f1, fmt=fmt1)
    t2 = trj.Trajectory(f2, fmt=fmt2)

    # Define slice.
    # We interpret --first N --last N as a request of step N
    if args.last == args.first and args.last is not None:
        args.last += 1
    sl1 = fractional_slice(args.first, args.last, args.skip, len(t1))
    sl2 = fractional_slice(args.first, args.last, args.skip, len(t2))

    # Here we could you a trajectory slice t[sl] but this will load
    # everything in ram (getitem doesnt provide a generator). This
    # will be fixed with python 3.
    ts1 = trajectory.Sliced(t1, sl1)
    ts2 = trajectory.Sliced(t2, sl2)

    def array_fmt(arr):
        """Remove commas and [] from numpy array repr."""
        # Passing a scalar will trigger an error (gotcha: even
        # when casting numpy array to list, the elements remain of
        # numpy type and this function gets called! (4% slowdown)
        _fmt = '%g'
        try:
            return ' '.join([_fmt % x for x in arr])
        except:
            return _fmt % arr
            # except:
            #     return numpy.array2string(arr, precision=self.precision, separator=',')[1:-1]
    import numpy
    numpy.set_string_function(array_fmt, repr=False)

    for step, s1, s2 in trj.utils.paste(ts1, ts2):
        try:
            for i in range(len(s1.particle)):
                print(getattr(s1.particle[i], attr1),
                      getattr(s2.particle[i], attr2))
        except:
            print(getattr(s1, attr1), getattr(s2, attr2))

def scatter(args):
    """
    Write frames in trajectory to individual files of the same file format
    """
    from atooms import trajectory as trj
    from atooms.core.utils import mkdir

    for f in args.file_inp:
        fmt = args.inp
        fmt_out = fmt
        t = trj.Trajectory(f, fmt=fmt)

        # Define slice
        # We interpret --first N --last N as a request of step N
        if args.last == args.first and args.last is not None:
            args.last += 1
        sl = fractional_slice(args.first, args.last, args.skip, len(t))
        ts = trajectory.Sliced(t, sl)
        for i, system in enumerate(ts):
            f_out = args.file_out.format(step=ts.steps[i], frame=i,
                                         base=os.path.splitext(f)[0],
                                         ext=os.path.splitext(f)[1])
            mkdir(os.path.dirname(f_out))
            with trj.Trajectory(f_out, fmt=fmt_out, mode='w') as th_out:
                if args.fields is not None:
                    th_out.variables = args.fields.split(',')
                else:
                    th_out.variables = ts.variables
                th_out.write(system, step=ts.steps[i])


if __name__ == '__main__':

    # create the top-level parser
    parser = argparse.ArgumentParser(epilog=formats(),
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser = add_first_last_skip(parser)
    parser.add_argument('-v', '--verbose', dest='verbose', action='store_true', help='verbose output')
    parser.set_defaults(func=lambda args: parser.print_help())
    subparsers = parser.add_subparsers()

    parser_convert = subparsers.add_parser('convert')
    parser_convert.add_argument(      '--fields', dest='fields', help='attributes fields')
    parser_convert.add_argument('-I', '--fields-include', dest='fields_include', type=str, default='', help='include patterns in fields')
    parser_convert.add_argument('-E', '--fields-exclude', dest='fields_exclude', type=str, default='', help='exclude patterns from fields')
    parser_convert.add_argument('-i', '--fmt-inp', dest='inp', help='input format')
    parser_convert.add_argument('-o', '--fmt-out', dest='out', help='output format for conversion')
    parser_convert.add_argument(      '--folder', dest='folder', action='store_true', help='force folder-based layout')
    parser_convert.add_argument('-F', '--ff', dest='ff', type=str, default='', help='force field file')
    parser_convert.add_argument(      '--flatten-steps',dest='flatten_steps', action='store_true', help='use sample index instead of steps')
    parser_convert.add_argument(      '--unfold',dest='unfold', action='store_true', help='unfold')
    parser_convert.add_argument(      '--fix-cm',dest='fix_cm', action='store_true', help='fix cm')
    parser_convert.add_argument(      '--side', dest='side', default=None, help='set cell side')
    parser_convert.add_argument(      '--density', dest='rho', type=float, default=None, help='new density')
    parser_convert.add_argument('-T', '--temperature', dest='temperature', type=float, default=None, help='new temperature')
    parser_convert.add_argument('-N', '--number-of-particles', dest='N', type=int, default=-1, help='new number of particles')
    parser_convert.add_argument(      '--precision', dest='precision', type=int, default=None, help='write precision')
    parser_convert.add_argument(      '--species',dest='species_layout', default=None, help='modify species layout (A, C, F)')
    parser_convert.add_argument(      '--sort-species',dest='species_sort', action='store_true', help='sort by species')
    parser_convert.add_argument(      '--seed', dest='seed', type=int, help='set seed of random number generator')
    parser_convert.add_argument(nargs=1, dest='file_inp', default='-', help='input file')
    parser_convert.add_argument(nargs='?', dest='file_out', default='-', help='output file')
    parser_convert.set_defaults(func=main)

    parser_info = subparsers.add_parser('info')
    parser_info.add_argument(      '--folder', dest='folder', action='store_true', help='force folder-based layout')
    parser_info.add_argument(      '--what', dest='fields', help='what info to show')
    parser_info.add_argument('-i', '--fmt-inp', dest='inp', help='input format')
    parser_info.add_argument(nargs='*', dest='file_inp', default='-', help='input file')
    parser_info.set_defaults(func=main_info)

    parser_paste = subparsers.add_parser('paste')
    parser_paste.add_argument('-i', '--fmt-inp', dest='inp', help='input format')
    parser_paste.add_argument(nargs=2, dest='file_inp', help='input files')
    parser_paste.set_defaults(func=main_paste)

    parser_scatter = subparsers.add_parser('scatter')
    parser_scatter.add_argument(      '--fields', dest='fields', help='attributes fields')
    parser_scatter.add_argument('-i', '--fmt-inp', dest='inp', help='input format')
    parser_scatter.add_argument('-o', '--output-file', dest='file_out', default='{base}_{frame}{ext}', help='output path (interpolated)')
    parser_scatter.add_argument(nargs='*', dest='file_inp', help='input file')
    parser_scatter.set_defaults(func=scatter)

    # parse argument lists
    args = parser.parse_args()
    args.func(args)



# This file is part of atooms
# Copyright 2010-2024, Daniele Coslovich

"""
Dry run backend.

It just exposes a minimal backend interface.
"""

import copy
import numpy


class DryRun:

    """A simulation backend that performs no simulation at all"""

    version = '0.1.0'

    def __init__(self, system=None):
        self.system = system
        if self.system is None:
            self.system = System()
        self.trajectory_class = Trajectory
        self.steps = 0

    def __str__(self):
        return 'dryrun'

    def write_checkpoint(self, output_path):
        pass

    def read_checkpoint(self, output_path):
        pass

    @property
    def rmsd(self):
        return 0.0

    def run(self, steps):
        pass


class System:

    """
    The `System` class monkey patches the relevant methods of
    `atooms.system.System` required for a valid simulation backend.
    """

    def __init__(self, particle=None, cell=None, thermostat=None,
                 barostat=None, reservoir=None):
        self.particle = particle if particle is not None else []
        self.cell = None
        self.thermostat = Thermostat()
        self.barostat = None
        self.reservoir = None

    def potential_energy(self, per_particle=False, normed=False, cache=False):
        return 0.

    def kinetic_energy(self, per_particle=False, normed=False):
        return 0.

    def total_energy(self, per_particle=False, normed=False, cache=False):
        return 0.

    def center_of_mass(self, what):
        return numpy.zeros(3)

    @property
    def density(self):
        return 0.

    @property
    def temperature(self):
        return 0.

    def set_temperature(self, T):
        pass

    def scale_velocities(self, factor):
        for p in self.particle:
            p.velocity *= -1.0

    def dump(self, clear=False):
        pass

    def update(self, other, full=False, exclude=None, only=None):
        for key in other.__dict__:
            if exclude is not None or only is not None:
                if (exclude is not None and key not in exclude) or \
                   (only is not None and key in only):
                    self.__dict__[key] = copy.deepcopy(other.__dict__[key])
            else:
                if full or other.__dict__[key] is not None:
                    self.__dict__[key] = copy.deepcopy(other.__dict__[key])

    def __str__(self):
        return 'DryRun'


class Thermostat:

    """A place-holder for the system thermostat and its temperature."""

    def __init__(self):
        self.temperature = 1.0


class Trajectory:

    """A minimal trajectory class interface."""

    suffix = 'dry'

    def __init__(self, filename, mode='r'):
        pass

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def write(self, system, step):
        pass

    def close(self):
        pass


class EnergyMinimization:

    "An optimization backend that performs no optimization at all."

    version = '0.1.0'

    def __init__(self, system):
        self.system = system
        self.trajectory_class = Trajectory
        self.output_path = None
        self.method = 'cg'
        self.tolerance = 1e-10
        self.max_iterations = 100000

    def run(self):
        pass



import copy
import numpy
from atooms.system.particle import Particle, cm_position, _periodic_vector_unfolded
import re
# For the network models, like Keating, it does not make much sense to
# use Molecule actually: the system would be one big molecule. One
# could just abuse of the NeighborList so that it does not get updated
# during the simulation! This would also allow bond switches.

class Molecule:

    def __init__(self, particle, bond, angle=None, species=None, cell=None):
        # For consistency with System, we use singular names for variables:
        # particle, bond, etc.
        self.particle = copy.deepcopy(particle)
        self.bond = bond
        if angle is None:
            angle = []
        self.angle = angle
        self.dihedral = []
        self.improper = []
        self.species = species
        if species is None:
            self.species = ''.join([str(p.species) for p in self.particle])
        self.cell = cell
        # Fold particles in cell when using PBCs
        if self.cell:
            for p in self.particle:
                p.fold(self.cell)

    @property
    def center_of_mass(self):
        if self.cell is None:
            return cm_position(self.particle)
        particle = [self.particle[0].nearest_image_of(p, self.cell) for p in self.particle]
        cm = cm_position(particle)
        cm = _periodic_vector_unfolded(cm, self.cell.side)
        return cm

    @center_of_mass.setter
    def center_of_mass(self, position):
        position = numpy.array(position)
        cm = self.center_of_mass
        for p in self.particle:
            p.position += (position - cm)
        if self.cell:
            for p in self.particle:
                p.fold(self.cell)

    @property
    def orientation(self):
        """
        Get the current orientation of the particles in the molecule.

        The orientation is calculated as the displacement vectors of each particle
        relative to the center of mass. The calculation takes into account periodic boundaries
        if the cell information is provided.

        Returns:
            numpy.ndarray: An array of vectors representing the orientation of each particle
            relative to the center of mass.
        """
        o = []
        cm = self.center_of_mass
        for p in self.particle:
            rij = p.position - cm
            if self.cell:
                rij = _periodic_vector_unfolded(rij, self.cell.side)
            o.append(rij)
        return numpy.array(o)

    @orientation.setter
    def orientation(self, value):
        """
        Set the new orientation of the particles in the molecule.

        This method updates the positions of the particles by adding the provided
        orientation vectors to the current center of mass. The number of elements in
        the `value` must match the number of particles in the molecule.

        Args:
            value (numpy.ndarray): An array of new orientation vectors for each particle.
        """
        assert len(value) == len(self.particle), 'the number of orientations must match the number of particles'
        cm = self.center_of_mass
        for p, o in zip(self.particle, value):
            p.position[:] = cm + o

    def rotate(self, theta, axis=None):
        """
        Rotate the molecule by `theta`, around the given `axis` in 3D

        In 2D, the method applies a 2D rotation matrix to the molecule's orientation using
        the provided angle `theta`. The rotation is counterclockwise in the 2D plane.

        In 3D, the method applies a 3D rotation matrix to the molecule's orientation using
        Rodrigues' rotation formula. The axis of rotation must be provided as a 3D
        vector. The angle `theta` should be in radians.
        """
        if len(self.particle[0].position) == 2:
            self._rotate_2D(theta)
        elif len(self.particle[0].position) == 3:
            assert axis is not None, 'provide axis for rotation'
            self._rotate_3D(theta, axis=axis)
        else:
            raise ValueError('cannot rotate molecules in dimensions other than 2 or 3')

    def _rotate_2D(self, theta):
        """
        Rotate the 2D molecule by a given angle.

        This method applies a 2D rotation matrix to the molecule's orientation using
        the provided angle `theta`. The rotation is counterclockwise in the 2D plane.

        Args:
            theta (float): The angle of rotation in radians.
        """
        rotation_matrix = numpy.array([[numpy.cos(theta), -numpy.sin(theta)],
                                       [numpy.sin(theta), numpy.cos(theta)]])
        self._apply_rotation_matrix(rotation_matrix)

    def _rotate_3D(self, theta, axis):
        """
        Rotate the molecule in 3D by a given angle around a specified axis.

        This method applies a 3D rotation matrix to the molecule's orientation using
        Rodrigues' rotation formula. The axis of rotation must be provided as a 3D
        vector. The angle `theta` should be in radians.

        Args:
            theta (float): The angle of rotation in radians.
            axis (array-like or str): A 3D vector representing the axis of rotation.
            If a string is provided, it must be compatible with `orientation_vector()`
        """
        axis = numpy.array(axis)
        # Since orientation_vector accepts a string now, things are simpler
        # if numpy.issubdtype(axis.dtype, numpy.str_) or numpy.issubdtype(axis.dtype, numpy.object_):
        if isinstance(axis, str):
            axis = self.orientation_vector(axis)
        axis /= numpy.linalg.norm(axis)  # Ensure u is a unit vector

        ux, uy, uz = axis
        cos_theta = numpy.cos(theta)
        sin_theta = numpy.sin(theta)

        # Skew-symmetric cross-product matrix K
        k = numpy.array([[  0,   -uz,   uy],
                         [ uz,    0,  -ux],
                         [-uy,   ux,    0]])

        # Rodrigues' formula: R = I + sin(theta) * K + (1 - cos(theta)) * K^2
        rotation_matrix = numpy.eye(3) + sin_theta * k + (1 - cos_theta) * numpy.dot(k, k)
        self._apply_rotation_matrix(rotation_matrix)

    def _apply_rotation_matrix(self, rotation_matrix):
        """
        Apply a rotation matrix to the molecule's orientation.

        This method updates the molecule's orientation by applying the provided
        rotation matrix. The new orientation is calculated by multiplying the current
        orientation by the transpose of the rotation matrix.

        Args:
            rotation_matrix (numpy.ndarray): A 2x2 or 3x3 rotation matrix to apply to the orientation.
        """
        self.orientation = numpy.dot(self.orientation, rotation_matrix.T)

    def orientation_vector(self, orientation, normed=False):
        """
        Compute the orientation vector based on `orientation` string.
        Orientation strings can be :
        - `etoe` (end-to-end),
        - 'i-j' (vector from particle `i` to `j`, or `CM`, `i`, `j` from 1 to N)
        - 'i-jxk-l' (cross product of two vectors)

        Args:
            orientation: string specifying the orientation vector.
            normed (bool): orientation becomes unitary if True
        Returns:
            numpy.ndarray: Array of orientation vector.
        """
        s = orientation
        # TODO: eat these delicious spaghetti
        if s in ['etoe', 'end-to-end', 'e2e']:
            rij = self._extract_endtoend()
        else: 
            rij = self._check_and_extract_vector(s)
            if rij is None:
                rij = self._check_and_extract_cross_product(s)
                if rij is None:
                    raise ValueError(f"Error: {s} is not a valid orientation")
            if self.cell:
                rij = _periodic_vector_unfolded(rij, self.cell.side)
        if normed:
            rij /= numpy.linalg.norm(rij)
        return numpy.array(rij)

    def custom_orientation(self, orientation, normed=False):
        """
        Compute the molecule orientation based on custom `orientation` strings.
        Orientation strings must be compatible with `orientation_vector`.

        Args:
            orientation (list): List of strings specifying the orientation vectors.
            normed (bool): orientation becomes unitary if True
        Returns:
            numpy.ndarray: Array of orientation vectors.
        """
        o = []
        for s in orientation:
            o.append(self.orientation_vector(s, normed))
        return numpy.array(o)
    
    def _extract_endtoend(self):
        """
        Check if the string specifies an end-to-end orientation and extract the vector.

        Args:
            s (str): Orientation string.

        Returns:
            numpy.ndarray or None: End-to-end vector if applicable, otherwise None.
        """
        def _unfold(particle, side):
            L = side
            old = particle[0].position.copy()
            new_pos = [old.copy()]
            for p in particle[1:]:
                pos = p.position.copy()
                dif = pos - old
                dif = dif - numpy.rint(dif / L) * L
                old += dif
                new_pos.append(old.copy())
            return new_pos[-1]

        if self.cell:
            last = _unfold(self.particle, self.cell.side)
        else:
            last = self.particle[-1].position
        rij = last - self.particle[0].position
        return rij

    def _check_and_extract_vector(self, s):
        """
        Check if the string specifies a vector orientation and extract the vector.

        Args:
            s (str): Orientation string.

        Returns:
            numpy.ndarray or None: Vector if applicable, otherwise None.
        """
        M = len(str(len(self.particle)))
        string_to_match = rf"^(CM|\d{{1,{M}}})-(CM|\d{{1,{M}}})$"
        pattern = re.compile(string_to_match)
        match = pattern.match(s)
        if match:
            x, y = match.groups()
            return self._get_vector(x, y)
        return None

    def _check_and_extract_cross_product(self, s):
        """
        Check if the string specifies a cross product orientation and extract the vector.

        Args:
            s (str): Orientation string.

        Returns:
            numpy.ndarray or None: Cross product vector if applicable, otherwise None.
        """
        M = len(str(len(self.particle)))
        string_to_match = rf"^(CM|\d{{1,{M}}})-(CM|\d{{1,{M}}})x(CM|\d{{1,{M}}})-(CM|\d{{1,{M}}})$"
        pattern = re.compile(string_to_match)
        match = pattern.match(s)
        if match:
            x, y, w, z = match.groups()
            vector_1 = self._get_vector(x, y)
            vector_2 = self._get_vector(w, z)
            cross_product = numpy.cross(vector_1, vector_2)
            if numpy.linalg.norm(cross_product) == 0:
                raise ValueError(f"Error: the cross product of {s} is zero. The orientation is not defined.")
            return cross_product
        return None

    def _get_vector(self, x, y):
        """
        Get the vector between two points specified by their indices or 'CM' (center of mass).

        Args:
            x (str): Index of the first point or 'CM' for center of mass.
            y (str): Index of the second point or 'CM' for center of mass.

        Returns:
            numpy.ndarray: Vector from point x to point y.

        Raises:
            ValueError: If x or y are not valid particle indices or 'CM'.
        """
        def get_position(index):
            return self.center_of_mass if index == "CM" else self.particle[int(index)-1].position

        try:
            position_x = get_position(x)
            position_y = get_position(y)
            return position_y - position_x
        except (ValueError, IndexError):
            raise ValueError(f"Error: {x if not x.isdigit() or int(x) >= self.len_molecule else y} is not a valid particle index,"
                             "ie 'CM' or an integer between 1 and {self.len_molecule}")



#!/usr/bin/env python

import os
import unittest
from atooms.trajectory import TrajectoryLAMMPS, MolecularTrajectoryLAMMPS
from atooms.trajectory.decorators import _Molecular


class Test(unittest.TestCase):

    def setUp(self):
        """
        Set up the test case by initializing the particle trajectory.
        """
        self.input_file = os.path.join(os.path.dirname(__file__),
                                       '../data/trimer_rho1.2.lammpstrj')
        self.particle_trajectory = TrajectoryLAMMPS(self.input_file)

    def test_trajectory(self):
        """
        Test the creation of the molecular trajectory from the particle trajectory.
        """
        molecular_trajectory = MolecularTrajectoryLAMMPS(self.input_file)
        self.frame_0 = molecular_trajectory[0]
        self.molecule_0 = self.frame_0.molecule[0]
        self.particle_0 = self.frame_0.particle[0]
        self.assertIsNotNone(self.molecule_0)
        self.assertIsNotNone(self.particle_0)
        molecular_trajectory.close()

    def test_decorator(self):
        MolecularTrajectory = _Molecular(TrajectoryLAMMPS)
        molecular_trajectory = MolecularTrajectory(self.input_file)
        self.assertTrue(isinstance(molecular_trajectory, TrajectoryLAMMPS))
        self.assertTrue(isinstance(molecular_trajectory, MolecularTrajectory))
        self.frame_0 = molecular_trajectory[0]
        self.molecule_0 = self.frame_0.molecule[0]
        self.particle_0 = self.frame_0.particle[0]
        self.assertIsNotNone(self.molecule_0)
        self.assertIsNotNone(self.particle_0)
        molecular_trajectory.close()

    def test_center_of_mass(self):
        """
        Test the calculation of the center of mass of the first molecule.
        """
        self.test_trajectory()  # Ensure test_trajectory runs first
        center_of_mass = self.molecule_0.center_of_mass
        self.assertAlmostEqual(center_of_mass[0], 3.129313, places=6)
        self.assertAlmostEqual(center_of_mass[1], 6.057427, places=6)
        self.assertAlmostEqual(center_of_mass[2], 4.262863, places=6)

    def test_position(self):
        """
        Test the position of the first particle in the first frame.
        """
        self.test_trajectory()  # Ensure test_trajectory runs first
        particle_0_position = self.particle_0.position
        self.assertAlmostEqual(particle_0_position[0], 2.96049, places=6)
        self.assertAlmostEqual(particle_0_position[1], 6.62826, places=6)
        self.assertAlmostEqual(particle_0_position[2], 4.19592, places=6)

    def _test_trajectory(self, cls):
        from atooms.core.utils import rmf

        with MolecularTrajectoryLAMMPS(self.input_file) as mth:
            s = mth[0]
        N_mol = len(s.molecule)
        fout = '/tmp/molecular.xyz'
        with _Molecular(cls)(fout, 'w') as mth:
            mth.write(s, 0)
        with _Molecular(cls)(fout) as mth:
            s = mth[0]
        self.assertEqual(len(s.molecule), N_mol)
        rmf(fout)

    def test_trajectory_any(self):
        from atooms.trajectory import TrajectoryXYZ, TrajectoryEXYZ, TrajectoryRam
        self._test_trajectory(TrajectoryXYZ)
        # TODO: fix warning about unclosed file, dont know why
        self._test_trajectory(TrajectoryEXYZ)
        # self._test_trajectory(TrajectoryRam)

    def test_system(self):
        """
        Test the length of the molecule list and particle list
        """
        self.test_trajectory()  # Ensure test_trajectory runs first
        self.assertEqual(len(self.frame_0.molecule), 1000)
        self.assertEqual(len(self.frame_0.particle), 3000)

    def test_unfold(self):
        import numpy
        from atooms.system import Particle, Molecule, System, Cell
        from atooms.trajectory import MolecularTrajectoryXYZ, Unfolded

        finp = '/tmp/test_molecular.xyz'
        molecule = Molecule([Particle(position=[2.9, 2.9], species=1),
                             Particle(position=[2.8, 2.8], species=2)], bond=[0, 1])
        s = System(molecule=[molecule], cell=Cell([6.0, 6.0]))
        with MolecularTrajectoryXYZ(finp, 'w') as th:
            th.write(s, 0)
            # This is like entering from the other side
            s.molecule[0].center_of_mass += numpy.array([-5.8, 0.0])
            th.write(s, 1)
        with Unfolded(MolecularTrajectoryXYZ(finp)) as th:
            self.assertEqual(th[1].molecule[0].particle[0].position[0], 3.1)
            self.assertEqual(th[1].molecule[0].particle[0].position[1], 2.9)
            self.assertAlmostEqual(th[1].molecule[0].orientation[0][0], th[0].molecule[0].orientation[0][0])
            self.assertAlmostEqual(th[1].molecule[0].orientation[0][1], th[0].molecule[0].orientation[0][1])

    def tearDown(self):
        self.particle_trajectory.close()


if __name__ == '__main__':
    unittest.main()
